from getpass import getpass
from os import getenv
from pathlib import Path
from typing import List

import click
from delfino.constants import PackageManager
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header
from delfino.utils import ArgsList
from delfino.validation import assert_package_manager_is_known, assert_pip_package_installed, pyproject_toml_key_missing

from delfino_core.config import CorePluginConfig, Dockerhub, pass_plugin_app_context

try:
    from packaging.version import Version
except ImportError:
    pass


def _install_emulators(build_for_platforms: List[str]) -> None:
    """See https://github.com/tonistiigi/binfmt#installing-emulators."""
    emulators = []
    if "linux/arm64" in build_for_platforms:
        emulators.append("arm64")
    if "linux/arm/v7" in build_for_platforms:
        emulators.append("arm")

    if emulators:
        print_header("Installing emulators", level=2, icon="⬇")
        run(
            ["docker", "run", "--privileged", "--rm", "tonistiigi/binfmt", "--install", *emulators],
            on_error=OnError.EXIT,
        )


def _docker_build(
    project_name: str,
    dockerhub: Dockerhub,
    flags: ArgsList,
    platform: str,
    tag: str = "latest",
    push: bool = False,
):
    _flags = list(flags)
    _flags.extend(["--platform", platform])
    if tag != "latest" and push:
        _flags.extend(["--cache-to", f"type=registry,ref={dockerhub.username}/{project_name}"])

    run(
        [
            "docker",
            "buildx",
            "build",
            "--cache-from",
            f"type=registry,ref={dockerhub.username}/{project_name}",
            "--tag",
            f"{dockerhub.username}/{project_name}:{tag}",
            *_flags,
            ".",
        ],
        on_error=OnError.EXIT,
    )


@click.command()
@click.option(
    "--push",
    is_flag=True,
    help="Push the built image and cache to remote docker image registry. Login is "
    "required and Personal Access Token will be requested from STDIN if "
    "`DOCKERHUB_PERSONAL_ACCESS_TOKEN` environment variable is not set.",
)
@click.option("--serialized", is_flag=True, help="Do not build multiple platforms concurrently.")
@pass_plugin_app_context
def build_docker(app_context: AppContext[CorePluginConfig], push: bool, serialized: bool):
    """Build and push a docker image."""
    plugin_config = app_context.plugin_config
    dockerfile = Path("Dockerfile")

    assert_package_manager_is_known(app_context.package_manager)
    assert (
        app_context.package_manager == PackageManager.POETRY
    ), f"Only the '{PackageManager.POETRY.value}' package manager is supported by this command."
    assert_pip_package_installed("packaging")
    assert dockerfile.exists() and dockerfile.is_file(), "File 'Dockerfile' not found but required by this command."
    assert app_context.pyproject_toml.tool.poetry, pyproject_toml_key_missing("tool.poetry")
    assert plugin_config.dockerhub, pyproject_toml_key_missing("tool.delfino.dockerhub")

    print_header("Running Docker build")

    flags: ArgsList = []

    poetry = app_context.pyproject_toml.tool.poetry
    project_name = poetry.name
    project_version = poetry.version

    python_version = Version(poetry.dependencies["python"].strip("<>=~^"))
    flags.extend(["--build-arg", f"PYTHON_VERSION={python_version.major}.{python_version.minor}"])

    dockerhub = plugin_config.dockerhub

    if push:
        dockerhub_password = getenv("DOCKERHUB_PERSONAL_ACCESS_TOKEN")
        if not dockerhub_password:
            click.secho("⚠ The 'DOCKERHUB_PERSONAL_ACCESS_TOKEN' environment variable is not set.", fg="yellow")
            dockerhub_password = getpass("Dockerhub Personal Access Token: ")
        run(f"docker login --username {dockerhub.username} --password {dockerhub_password}", on_error=OnError.EXIT)

    if getenv("CI"):  # https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables
        flags.extend(["--progress", "plain"])

    joined_build_platforms = ",".join(dockerhub.build_for_platforms)
    if serialized:
        build_platforms: List[str] = dockerhub.build_for_platforms
    else:
        build_platforms = [joined_build_platforms]
        _install_emulators(dockerhub.build_for_platforms)

    # If serialized build is selected, run build individually. Results are cached so the second
    # build below will only push and not build again.
    for platform in build_platforms:
        if serialized:
            _install_emulators([platform])
        print_header(f"Build {dockerhub.username}/{project_name}:latest for {platform}", level=2, icon="🔨")
        _docker_build(project_name, dockerhub, flags, platform)

    if not push:
        return

    flags.extend(["--output", "type=image,push=true"])

    # While `docker buildx build` supports multiple `--tag` flags, push of them fails to expose
    # all architectures in `latest`. Multiple pushes fix this.
    for tag in [project_version, "latest"]:
        print_header(f"Push {dockerhub.username}/{project_name}:{tag}", level=2, icon="🔨")
        _docker_build(project_name, dockerhub, flags, joined_build_platforms, tag, push)

    run("docker logout", on_error=OnError.EXIT)



from pathlib import Path
from subprocess import PIPE, CompletedProcess
from typing import Tuple

import click
from delfino.decorators import files_folders_option
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header, run_command_example
from delfino.validation import assert_pip_package_installed

from delfino_core.config import CorePluginConfig, pass_plugin_app_context


def _check_result(app_context: AppContext[CorePluginConfig], result: CompletedProcess, check: bool, msg: str):
    if result.returncode == 1 and check:

        msg_lines = [
            f"{msg} before commit. Try following:",
            f" * Run formatter manually with `{run_command_example(run_format, app_context)}` before committing code.",
        ]
        if not app_context.plugin_config.disable_pre_commit:
            msg_lines.insert(1, " * Enable pre-commit hook by running `pre-commit install` in the repository.")

        click.secho(
            "\n".join(msg_lines),
            fg="red",
            err=True,
        )
        raise click.Abort()

    if result.returncode > 1:
        raise click.Abort()


@click.command("format")
@click.option("--check", is_flag=True, help="Only check formatting, don't reformat the code.")
@click.option("--quiet", is_flag=True, help="Don't show progress. Only errors.")
@files_folders_option
@pass_plugin_app_context
def run_format(app_context: AppContext[CorePluginConfig], files_folders: Tuple[Path, ...], check: bool, quiet: bool):
    """Runs black code formatter and isort on source code."""
    plugin_config = app_context.plugin_config

    assert_pip_package_installed("isort")
    assert_pip_package_installed("black")

    if not plugin_config.disable_pre_commit:
        assert_pip_package_installed("pre-commit")
        # ensure pre-commit is installed
        run("pre-commit install", stdout=PIPE, on_error=OnError.EXIT)

    if not files_folders:
        files_folders = (plugin_config.sources_directory, plugin_config.tests_directory)
        if app_context.pyproject_toml.tool.delfino.local_commands_directory.exists():
            files_folders += (app_context.pyproject_toml.tool.delfino.local_commands_directory,)

    flags = []

    if check:
        flags.append("--check")

    print_header("Sorting imports", icon="ℹ")

    _check_result(
        app_context, run(["isort", *files_folders, *flags], on_error=OnError.PASS), check, "Import were not sorted"
    )

    print_header("Formatting code", icon="🖤")

    if quiet:
        flags.append("--quiet")

    _check_result(
        app_context, run(["black", *files_folders, *flags], on_error=OnError.PASS), check, "Code was not formatted"
    )



"""Type checking on source code."""

from itertools import groupby
from pathlib import Path
from typing import List, Tuple

import click
from delfino.decorators import files_folders_option, pass_args
from delfino.execution import OnError, run
from delfino.models import AppContext
from delfino.terminal_output import print_header
from delfino.utils import ArgsList
from delfino.validation import assert_pip_package_installed

from delfino_core.config import CorePluginConfig, pass_plugin_app_context
from delfino_core.utils import ensure_reports_dir


def _run_typecheck(
    paths: List[Path],
    strict: bool,
    reports_file: Path,
    summary_only: bool,
    mypypath: Path,
    print_second_level_headers: bool,
    passed_args: Tuple[str, ...],
):
    args: ArgsList = [
        "mypy",
        "--show-column-numbers",
        "--show-error-codes",
        "--color-output",
        "--warn-unused-config",
        "--warn-unused-ignores",
        "--color-output",
        "--allow-untyped-decorators",
        "--follow-imports",
        "silent",
        "--junit-xml",
        reports_file,
        *passed_args,
    ]

    if strict:
        args.append("--strict")

    args.extend(paths)

    if summary_only:
        args.extend(["|", "tail", "-n", "1"])

    if print_second_level_headers:
        print_header(f"{'Strict' if strict else 'Optional'} types", level=2, icon="⚠️" if strict else "ℹ️")

    run(args, env_update_path={"MYPYPATH": mypypath}, on_error=OnError.ABORT)


def is_path_relative_to_paths(path: Path, paths: List[Path]) -> bool:
    for _path in paths:
        try:
            path.relative_to(_path)
            return True
        except ValueError:
            continue
    return False


@click.command()
@click.option("--summary-only", is_flag=True, help="Suppress error messages and show only summary error count.")
@files_folders_option
@pass_args
@pass_plugin_app_context
def typecheck(
    app_context: AppContext[CorePluginConfig],
    passed_args: Tuple[str, ...],
    summary_only: bool,
    files_folders: Tuple[str, ...],
):
    """Run type checking on source code.

    A non-zero return code from this task indicates invalid types were discovered.
    """
    assert_pip_package_installed("mypy")

    print_header("RUNNING TYPE CHECKER", icon="🔠")

    plugin_config = app_context.plugin_config
    ensure_reports_dir(plugin_config)

    if files_folders:
        target_paths = [Path(path) for path in files_folders]
    else:
        target_paths = [plugin_config.sources_directory, plugin_config.tests_directory]
        if app_context.pyproject_toml.tool.delfino.local_commands_directory.exists():
            target_paths.append(app_context.pyproject_toml.tool.delfino.local_commands_directory)

    strict_paths = plugin_config.typecheck.strict_directories
    grouped_paths = groupby(target_paths, lambda current_path: is_path_relative_to_paths(current_path, strict_paths))

    for force_typing, group in grouped_paths:
        report_filepath = (
            plugin_config.reports_directory / "typecheck" / f"junit-{'strict' if force_typing else 'nonstrict'}.xml"
        )
        _run_typecheck(
            list(group),
            force_typing,
            report_filepath,
            summary_only,
            plugin_config.sources_directory,
            bool(strict_paths),
            passed_args,
        )
